#include <pthread.h>
#include <termios.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <linux/input.h>
#include "fsayskeyboard.h"

FSaysKeyboard* object_which_will_handle_signal;
FSaysKeyboard* object_which_will_process_keys;

void fsaysKeyboardWrapper(int t)
{
  object_which_will_handle_signal->alarm_func(t);
}

void* fsaysProcessKeyWrapper(void* p )
{
  object_which_will_process_keys = (FSaysKeyboard*) p; 
  object_which_will_process_keys->processKeys();

  return NULL;
}


FSaysKeyboard::FSaysKeyboard(char* eventPath){
    _p_firstKeystroke = NULL;
    _p_lastKeystroke =NULL;

    fd = open("/dev/input/event0", O_RDWR);
    //fd = open(eventPath, O_RDWR);
    if(fd){
        memset(&event, 0, sizeof(event));
        memset(&event, 0, sizeof(event_end));
        keepRunning = true;
        status = 0;
    }
    else
    {
        printf("Errro open keyboard:%s\n", strerror(errno));
        status = -1;
    }
    
}

FSaysKeyboard::~FSaysKeyboard()
{
    while(_p_firstKeystroke != NULL)
    {
        FSaysKeystroke* p_keystroke = _p_firstKeystroke;
        _p_firstKeystroke =_p_firstKeystroke->getNext();
        delete p_keystroke;
    }
    if(_p_currentRemoteKey != NULL)
    {
        delete _p_currentRemoteKey;
    }
    if(_p_previousRemoteKey != NULL)
    {
        delete _p_previousRemoteKey;
    }
}

int FSaysKeyboard::getStatus(void)
{
    return status;
}

int FSaysKeyboard::getch(void)
{
    struct termios oldattr, newattr;
    int ch;
    tcgetattr( STDIN_FILENO, &oldattr );
    newattr = oldattr;
    newattr.c_lflag &= ~( ICANON | ECHO );
    tcsetattr( STDIN_FILENO, TCSANOW, &newattr );
    ch = getchar();
    tcsetattr( STDIN_FILENO, TCSANOW, &oldattr );
    return ch;
}

/* reads from keypress, echoes */
int FSaysKeyboard::getche(void)
{
    struct termios oldattr, newattr;
    int ch;
    tcgetattr( STDIN_FILENO, &oldattr );
    newattr = oldattr;
    newattr.c_lflag &= ~( ICANON );
    tcsetattr( STDIN_FILENO, TCSANOW, &newattr );
    ch = getchar();
    tcsetattr( STDIN_FILENO, TCSANOW, &oldattr );
    return ch;
}
void FSaysKeyboard::startKey()
{
    if(_p_currentRemoteKey == NULL)
    {
        return;
    }
    int key = _p_currentRemoteKey->getKey();
    bool shiftKey = _p_currentRemoteKey->getShiftKey();

    //printf("startKey: %d, shift:%d\n", key, shiftKey);
    if(shiftKey)
    {
        gettimeofday(&event.time, NULL);
        event.type = EV_KEY;
        event.code = KEY_LEFTSHIFT;
        event.value = 1;
        write(fd, &event, sizeof(event));
    }
    gettimeofday(&event.time, NULL);
    event.type = EV_KEY;
    event.code = key;
    event.value = 1;
    usleep(MY_HSEC);
    gettimeofday(&event_end.time, NULL);
    event_end.type = EV_SYN;
    event_end.code = SYN_REPORT;
    event_end.value = 1;
    write(fd, &event, sizeof(event));
    write(fd, &event_end, sizeof(event_end));
    usleep(MY_FHSEC);

    _p_previousRemoteKey = _p_currentRemoteKey;
    _p_currentRemoteKey = new FSaysRemoteKey();
}

void FSaysKeyboard::endKey()
{
    if(_p_previousRemoteKey == NULL)
    {
        return;
    }
    int key = _p_previousRemoteKey->getKey();
    bool shiftKey = _p_previousRemoteKey->getShiftKey();

    //printf("endKey: %d, shift:%d\n", key, shiftKey);
    gettimeofday(&event.time, NULL);
    event.type = EV_KEY;
    event.code = key;
    event.value = 0;
    usleep(MY_HSEC);
    write(fd, &event, sizeof(event));
    if(shiftKey)
    {
        gettimeofday(&event.time, NULL);
        event.type = EV_KEY;
        event.code = KEY_LEFTSHIFT;
        event.value = 0;
        write(fd, &event, sizeof(event));
    }
    gettimeofday(&event_end.time, NULL);
    event_end.type = EV_SYN;
    event_end.code = SYN_REPORT;
    event_end.value = 0;
    write(fd, &event, sizeof(event));
    write(fd, &event_end, sizeof(event_end));
    usleep(MY_FHSEC);

    delete _p_previousRemoteKey;
    _p_previousRemoteKey= NULL;
}

void FSaysKeyboard::alarm_func(int t)
{
    sem_wait(&_mutex);
    if(_p_lastKeystroke != NULL)
    {
        _p_lastKeystroke->setTimedOut(true);
    }

    endKey();
    sem_post(&_mutex); 
}


void FSaysKeyboard::foo(){
    struct sigaction act;
    act.sa_handler = fsaysKeyboardWrapper; 
    object_which_will_handle_signal = this;    


    pthread_t thread_a;
    sem_init(&_mutex, 0, 1);
    pthread_create (&thread_a, NULL,  &fsaysProcessKeyWrapper, (void*) this);

    do 
    {
        sigaction(SIGALRM, &act,0);
        ualarm(MY_TSEC, 0);
        int key=getch();
        ualarm(0,0);
        FSaysKeystroke* p_keystroke = new FSaysKeystroke(key);
        sem_wait(&_mutex);
        //printf("enter critical_main\n");
        if(_p_firstKeystroke == NULL)
        {
            _p_firstKeystroke = p_keystroke;
        }
        if(_p_lastKeystroke != NULL)
        {
            _p_lastKeystroke->setNext(p_keystroke);
        }
        _p_lastKeystroke = p_keystroke;
        //printf("leaving critical_main\n");
        sem_post(&_mutex); 
    } while(keepRunning);


    close(fd);
    sem_destroy(&_mutex);
}

// PRE: first key stroke is esape
bool FSaysKeyboard::handleEscape()
{
    bool foundMatchForKeyStrokes = false;
    int key1 = _p_firstKeystroke->getKey();
    int key2 = -1; 
    int key3 = -1; 
    int key4 = -1; 
    int key5 = -1; 
    FSaysKeystroke* p_secondKeystroke = _p_firstKeystroke->getNext();
    FSaysKeystroke* p_thirdKeystroke = NULL;

    if(p_secondKeystroke != NULL)
    {
        key2 = p_secondKeystroke->getKey();    
        p_thirdKeystroke = p_secondKeystroke->getNext(); 
    }
    if(p_secondKeystroke != NULL && p_thirdKeystroke != NULL)
    {
        key3 = p_thirdKeystroke->getKey();    
        
        foundMatchForKeyStrokes = handleEscape3chars(key2, key3);
        if(foundMatchForKeyStrokes)
        {
            deleteNkeyStrokes(3);
        }
        else
        {
            FSaysKeystroke* p_fourthKeystroke = p_fourthKeystroke = p_thirdKeystroke->getNext();
            if(p_fourthKeystroke != NULL)
            {
                key4 = p_fourthKeystroke->getKey();    
                foundMatchForKeyStrokes = handleEscape4chars(key2, key3, key4);
                if(foundMatchForKeyStrokes)
                {
                    deleteNkeyStrokes(4);
                }
                else
                {
                    FSaysKeystroke* p_fifthKeystroke = p_fifthKeystroke = p_fourthKeystroke->getNext();
                    if(p_fifthKeystroke != NULL)
                    {
                        key5 = p_fifthKeystroke->getKey();    
                        foundMatchForKeyStrokes = handleEscape5chars(key2, key3, key4, key5);
                    
                        if(!foundMatchForKeyStrokes)
                        {
                            printf("Unhandled key sequence, cleaning up...\n");
                        }
                        deleteNkeyStrokes(5);
                    }
                }
            }
        }
    }
    
    return foundMatchForKeyStrokes;
}
bool FSaysKeyboard::handleEscape3chars(int key2, int key3)
{
    bool foundMatchForKeyStrokes = false;
    if(key2 == 79)
    {
        foundMatchForKeyStrokes = true;
        switch(key3)
        {
            case 80:
                _p_currentRemoteKey->setKey(KEY_F1);
                printf("[F1]\n");
            break;
            case 81:
                _p_currentRemoteKey->setKey(KEY_F2);
                printf("[F2]\n");
            break;
            case 82:
                _p_currentRemoteKey->setKey(KEY_F3);
                printf("[F3]\n");
            break;
            case 83:
                _p_currentRemoteKey->setKey(KEY_F4);
                printf("[F4]\n");
            break;
            default:
                //printf("[OOPS UNHANDLED 3 key escape, key2: %d, key3 %d]\n", key2, key3);
                foundMatchForKeyStrokes = false;
            break;
        }
    }
    else if(key2 == 91)
    {
        foundMatchForKeyStrokes = true;
        switch(key3)
        {
            case 65:
                _p_currentRemoteKey->setKey(KEY_UP);
                printf("[UP]\n");
            break;
            case 66:
                _p_currentRemoteKey->setKey(KEY_DOWN);
                printf("[DOWN]\n");
            break;
            case 67:
                _p_currentRemoteKey->setKey(KEY_RIGHT);
                printf("RIGHT]\n");
            break;
            case 68:
                _p_currentRemoteKey->setKey(KEY_LEFT);
                printf("[LEFT]\n");
            break;
            default:
                //printf("[OOPS UNHANDLED 3 key escape, key2: %d, key3%]\n", key2, key3);
                foundMatchForKeyStrokes = false;
            break;
        }
    }
    else if(key2 == 91)
    {
        //printf("[OOPS UNHANDLED 3 key escape, key2: %d, key3 %d]\n", key2, key3);
    }
    return foundMatchForKeyStrokes;
}

bool FSaysKeyboard::handleEscape4chars(int key2, int key3, int key4)
{
    bool foundMatchForKeyStrokes = false;
    printf("[ESC4]\n");
    if(key4 != 126)
    {
        return false;
    }
    if(key2 == 91)
    {
        foundMatchForKeyStrokes = true;
        switch(key3)
        {
            case 49:
                _p_currentRemoteKey->setKey(KEY_F1);
                printf("[F1]\n");
            break;
            case 50:
                _p_currentRemoteKey->setKey(KEY_INSERT);
                printf("[INSERT]\n");
            break;
            case 51:
                _p_currentRemoteKey->setKey(KEY_DELETE);
                printf("[DELETE]\n");
            break;
            case 52:
                _p_currentRemoteKey->setKey(KEY_END);
                printf("[END]\n");
            break;
            case 53:
                _p_currentRemoteKey->setKey(KEY_PAGEUP);
                printf("[PAGE UP]\n");
            break;
            case 54:
                _p_currentRemoteKey->setKey(KEY_PAGEDOWN);
                printf("[PAGE DOWN]\n");
            break;
            case 56:
                _p_currentRemoteKey->setKey(KEY_F7);
                printf("[F7]\n");
            break;
            default:
                printf("[OOPS UNHANDLED 4 key escape, key2: %d, key3: %d, key4, %d]\n", key2, key3, key4);
                foundMatchForKeyStrokes = false;
            break;
        }
    }
    return foundMatchForKeyStrokes;
}

// PRE: [0] == local escape , [4] == 126
bool FSaysKeyboard::handleEscape5chars(int key2, int key3, int key4, int key5)
{
    bool foundMatchForKeyStrokes = false;
    printf("[ESC5]\n");
    if(key5 != 126)
    {
        printf("[OOPS UNHANDLED 5 key escape, key2: %d, key3: %d, key4, %d, key5 %d]\n", key2, key3, key4, key5);
        return false;
    }
    if(key2 == 91 && key3 == 49)
    {
        foundMatchForKeyStrokes = true;
        switch(key4)
        {
            case 49:
                _p_currentRemoteKey->setKey(KEY_F1);
                printf("[F1]\n");
            break;
            case 50:
                _p_currentRemoteKey->setKey(KEY_F2);
                printf("[F2]\n");
            break;
            case 51:
                _p_currentRemoteKey->setKey(KEY_F3);
                printf("[F3]\n");
            break;
            case 52:
                _p_currentRemoteKey->setKey(KEY_F4);
                printf("[F4]\n");
            break;
            case 53:
                _p_currentRemoteKey->setKey(KEY_F5);
                printf("[F5]\n");
            break;
            case 55:
                _p_currentRemoteKey->setKey(KEY_F6);
                printf("[F6]\n");
            break;
            case 56:
                _p_currentRemoteKey->setKey(KEY_F7);
                printf("[F7]\n");
            break;
            case 57:
                _p_currentRemoteKey->setKey(KEY_F8);
                printf("[F8]\n");
            break;
            default:
                printf("[OOPS UNHANDLED 5 key escape, key2: %d, key3: %d, key4, %d, key5 %d]\n", key2, key3, key4, key5);
                foundMatchForKeyStrokes = false;
            break;
        }
    }
    else if(key2 == 91 && key3 == 50)
    {
        foundMatchForKeyStrokes = true;
        switch(key4)
        {
            case 48:
                _p_currentRemoteKey->setKey(KEY_F9);
                printf("[F9]\n");
            break;
            case 49:
                _p_currentRemoteKey->setKey(KEY_F10);
                printf("[F10]\n");
            break;
            case 51:
                _p_currentRemoteKey->setKey(KEY_F11);
                printf("[F11]\n");
            break;
            case 52:
                _p_currentRemoteKey->setKey(KEY_F12);
                printf("[F12]\n");
            break;
            default:
                printf("[OOPS UNHANDLED 5 key escape, key2: %d, key3: %d, key4, %d, key5 %d]\n", key2, key3, key4, key5);
                foundMatchForKeyStrokes = false;
            break;
        }
    }
    return foundMatchForKeyStrokes;
}

bool FSaysKeyboard::handleSingleKey(int key)
{
    bool foundMatchForKeyStrokes = false;
    int currentRemoteKey = -1;
    if(key >='a' && key <='z')
    {
        foundMatchForKeyStrokes = handleAZ(key);
    }
    else if(key >='A' && key <='Z')
    { 
        printf("[shift] + ");
        foundMatchForKeyStrokes = handleAZ(key);
        if(foundMatchForKeyStrokes)
        {
            _p_currentRemoteKey->setShiftKey(true);
        }
    }
    else 
    {   
        foundMatchForKeyStrokes = true;
        switch(key)
        {
        case 10:
            _p_currentRemoteKey->setKey(KEY_ENTER);
            printf("[ENTER]\n");
        break;
        case 127:
            _p_currentRemoteKey->setKey(KEY_BACKSPACE);
            printf("[BACKSPACE]\n");
        break;
        case ' ':
            _p_currentRemoteKey->setKey(KEY_SPACE);
            printf("[SPACE]\n");
        break;
        case '0':
            _p_currentRemoteKey->setKey(KEY_0);
            printf("[KEYPAD 0]\n");
        break;
        case '1':
            _p_currentRemoteKey->setKey(KEY_1);
            printf("[1]\n");
        break;
        case '2':
            _p_currentRemoteKey->setKey(KEY_2);
            printf("[2]\n");
        break;
        case '3':
            _p_currentRemoteKey->setKey(KEY_3);
            printf("[3]\n");
        break;
        case '4':
            _p_currentRemoteKey->setKey(KEY_4);
            printf("[4]\n");
        break;
        case '5':
            _p_currentRemoteKey->setKey(KEY_5);
            printf("[5]\n");
        break;
        case '6':
            _p_currentRemoteKey->setKey(KEY_6);
            printf("[6]\n");
        break;
        case '7':
            _p_currentRemoteKey->setKey(KEY_7);
            printf("[7]\n");
        break;
        case '8':
            _p_currentRemoteKey->setKey(KEY_8);
            printf("[8]\n");
        break;
        case '9':
            _p_currentRemoteKey->setKey(KEY_9);
            printf("[9]\n");
        break;
        case '-':
            _p_currentRemoteKey->setKey(KEY_MINUS);
            printf("[-]\n");
        break;
        case '=':
            _p_currentRemoteKey->setKey(KEY_EQUAL);
            printf("[=]\n");
        break;
        case '\t':
            _p_currentRemoteKey->setKey(KEY_TAB);
            printf("[\t]\n");
        break;
        case '[':
            _p_currentRemoteKey->setKey(KEY_LEFTBRACE);
            printf("[\n");
        break;
        case ']':
            _p_currentRemoteKey->setKey(KEY_RIGHTBRACE);
            printf("]\n");
        break;
        case ';':
            _p_currentRemoteKey->setKey(KEY_SEMICOLON);
            printf("[;]\n");
        break;
        case '\'':
            _p_currentRemoteKey->setKey(KEY_APOSTROPHE);
            printf("[']\n");
        break;
        case '`':
            _p_currentRemoteKey->setKey(KEY_GRAVE);
            printf("[`]\n");
        break;
        case '\\':
            _p_currentRemoteKey->setKey(KEY_BACKSLASH);
            printf("[\\]\n");
        break;
        case ',':
            _p_currentRemoteKey->setKey(KEY_COMMA);
            printf("[,]\n");
        break;
        case '.':
            _p_currentRemoteKey->setKey(KEY_DOT);
            printf("[.]\n");
        break;
        case '/':
            _p_currentRemoteKey->setKey(KEY_SLASH);
            printf("[/]\n");
        break;
        case '*':
            _p_currentRemoteKey->setKey(KEY_KPASTERISK);
            printf("[*]\n");
        break;
        default:
            printf("[OOPS UNHANDLED]\n");
            foundMatchForKeyStrokes = false;
        break;
        }
    }


    if(foundMatchForKeyStrokes)
    {   
        FSaysKeystroke* currentKeystroke = _p_firstKeystroke;
        _p_firstKeystroke = _p_firstKeystroke->getNext();
        if(_p_firstKeystroke == NULL)
        {
            _p_lastKeystroke = NULL;
        }
        delete currentKeystroke;
    }
    return foundMatchForKeyStrokes;
}

void FSaysKeyboard::deleteNkeyStrokes(int n)
{
    for(int i=0;i<n;i++)
    {
        FSaysKeystroke* currentKeystroke = _p_firstKeystroke;
        _p_firstKeystroke = _p_firstKeystroke->getNext();
        if(_p_firstKeystroke == NULL)
        {
            _p_lastKeystroke = NULL;
        }
          
        printf("Deleting keystroke #%d, %d\n", i+1, currentKeystroke);
        delete currentKeystroke;
    }
}


bool FSaysKeyboard::handleAZ(int key)
{
    switch(key)
    {
        case 'a':
        case 'A':
            _p_currentRemoteKey->setKey(KEY_A);
            printf("[a]\n");
        break;
        case 'b':
        case 'B':
            _p_currentRemoteKey->setKey(KEY_B);
            printf("[b]\n");
        break;
        case 'c':
        case 'C':
            _p_currentRemoteKey->setKey(KEY_C);
            printf("[c]\n");
        break;
        case 'd':
        case 'D':
            _p_currentRemoteKey->setKey(KEY_D);
            printf("[d]\n");
        break;
        case 'e':
        case 'E':
            _p_currentRemoteKey->setKey(KEY_E);
            printf("[e]\n");
        break;
        case 'f':
        case 'F':
            _p_currentRemoteKey->setKey(KEY_F);
            printf("[f]\n");
        break;
        case 'g':
        case 'G':
            _p_currentRemoteKey->setKey(KEY_G);
            printf("[g]\n");
        break;
        case 'h':
        case 'H':
            _p_currentRemoteKey->setKey(KEY_H);
            printf("[h]\n");
        break;
        case 'i':
        case 'I':
            _p_currentRemoteKey->setKey(KEY_I);
            printf("[i]\n");
        break;
        case 'j':
        case 'J':
            _p_currentRemoteKey->setKey(KEY_J);
            printf("[j]\n");
        break;
        case 'k':
        case 'K':
            _p_currentRemoteKey->setKey(KEY_K);
            printf("[k]\n");
        break;
        case 'l':
        case 'L':
            _p_currentRemoteKey->setKey(KEY_L);
            printf("[l]\n");
        break;
        case 'm':
        case 'M':
            _p_currentRemoteKey->setKey(KEY_M);
            printf("[m]\n");
        break;
        case 'n':
        case 'N':
            _p_currentRemoteKey->setKey(KEY_N);
            printf("[n]\n");
        break;
        case 'o':
        case 'O':
            _p_currentRemoteKey->setKey(KEY_O);
            printf("[o]\n");
        break;
        case 'p':
        case 'P':
            _p_currentRemoteKey->setKey(KEY_P);
            printf("[p]\n");
        break;
        case 'q':
        case 'Q':
            _p_currentRemoteKey->setKey(KEY_Q);
            printf("[q]\n");
        break;
        case 'r':
        case 'R':
            _p_currentRemoteKey->setKey(KEY_R);
            printf("[r]\n");
        break;
        case 's':
        case 'S':
            _p_currentRemoteKey->setKey(KEY_S);
            printf("[s]\n");
        break;
        case 't':
        case 'T':
            _p_currentRemoteKey->setKey(KEY_T);
            printf("[t]\n");
        break;
        case 'u':
        case 'U':
            _p_currentRemoteKey->setKey(KEY_U);
            printf("[u]\n");
        break;
        case 'v':
        case 'V':
            _p_currentRemoteKey->setKey(KEY_V);
            printf("[v]\n");
        break;
        case 'w':
        case 'W':
            _p_currentRemoteKey->setKey(KEY_W);
            printf("[w]\n");
        break;
        case 'x':
        case 'X':
            _p_currentRemoteKey->setKey(KEY_X);
            printf("[x]\n");
        break;
        case 'y':
        case 'Y':
            _p_currentRemoteKey->setKey(KEY_Y);
            printf("[y]\n");
        break;
        case 'z':
        case 'Z':
            _p_currentRemoteKey->setKey(KEY_Z);
            printf("[z]\n");
        break;
    }
    return true;
}

void FSaysKeyboard::processKeys(void)
{
    printf("background handler\n"); 
    _p_currentRemoteKey = new FSaysRemoteKey();

    while(keepRunning)
    {
        bool readQueue = true;
        sem_wait(&_mutex); 
        while(readQueue)
        {
            bool processRemoteKey = false;
            int currentRemoteKey = -1;
            //printf("enter critical_sub\n");
            if(_p_firstKeystroke != NULL)
            {
                int key = _p_firstKeystroke->getKey();
                if(key == LOCAL_ESCAPE)
                {
                    if (_p_firstKeystroke->getTimedOut())
                    {
                        printf("Escape: press enter to exit...");
                        keepRunning = false;
                    }
                    else
                    {
                        processRemoteKey = handleEscape();
                    }
                }
                else
                {
                    processRemoteKey = handleSingleKey(key);
                }
            }
            //printf("leaving critical_sub\n");
            if(processRemoteKey)
            {
                if(_p_previousRemoteKey != NULL)
                {
                    if(_p_previousRemoteKey->getKey() != _p_currentRemoteKey->getKey() || 
                    _p_previousRemoteKey->getShiftKey() != _p_currentRemoteKey->getShiftKey())
                    {
                        endKey();
                        startKey(); 
                    }
                    else
                    {
                        delete _p_previousRemoteKey;
                        _p_previousRemoteKey = _p_currentRemoteKey;
                        _p_currentRemoteKey = new FSaysRemoteKey();
                        printf("Same old, same old...");
                    }
                }
                else
                {
                    startKey(); 
                }
                //delete _p_currentRemoteKey;
                //_p_currentRemoteKey = new FSaysRemoteKey();
            }
            readQueue = processRemoteKey;
        }
        sem_post(&_mutex);  
        //sleep(1);
        usleep(MY_TSEC);
    }
}
